//This example ilustrates how to use object-oriented interface of the G39DDC API
//in non-object C language. It is recomended to use object-oriented interface
//instead of non-object interface in C because it is more efecient than non-object
//interface. Each function of non-object interface verifies device handle
//and translates the handle to reference to internal data structures, methods
//of object interface do not do it.

#include "G39DDCAPI.h"
#include <stdio.h>
#include <mmsystem.h>
#include <math.h>

#define MSG_EXITTHREAD (WM_APP+1000)
                                       
//prototypes of callback methods
VOID __stdcall MyCallback_G39DDC_DDC1StreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,CONST VOID *Buffer,UINT32 NumberOfSamples,UINT32 BitsPerSample);
BOOL __stdcall MyCallback_G39DDC_DDC1PlaybackStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,VOID *Buffer,UINT32 NumberOfSamples,UINT32 BitsPerSample);
VOID __stdcall MyCallback_G39DDC_DDC2StreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,CONST float *Buffer,UINT32 NumberOfSamples);
VOID __stdcall MyCallback_G39DDC_DDC2PreprocessedStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,CONST float *Buffer,UINT32 NumberOfSamples,float SlevelPeak,float SlevelRMS);
VOID __stdcall MyCallback_G39DDC_IFCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,CONST SHORT *Buffer,UINT32 NumberOfSamples,UINT32 CenterFrequency,WORD MaxADCAmplitude,UINT32 ADCSampleRate);
BOOL __stdcall MyCallback_G39DDC_SweepingCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,CONST SHORT *Buffer,UINT32 NumberOfSamples,UINT32 CenterFrequency,UINT64 FrontEndFrequency);
VOID __stdcall MyCallback_G39DDC_AudioStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,UINT32 Type,CONST float *Buffer,UINT32 NumberOfSamples);
BOOL __stdcall MyCallback_G39DDC_AudioPlaybackStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,float *Buffer,UINT32 NumberOfSamples);

IG39DDCDevice *Device;
HMODULE hAPI=NULL;
HANDLE hThread;
DWORD ThreadId;
HWAVEOUT hWaveOut;
LONG BufferCount;
G39DDCAPI_CREATE_INSTANCE CreateInstance;

//table of virtual methods
IG39DDCDeviceCallbackVtbl MyCallbackMethodsTable=
{
    //method order is important
    MyCallback_G39DDC_DDC1StreamCallback,
    MyCallback_G39DDC_DDC1PlaybackStreamCallback,
    MyCallback_G39DDC_DDC2StreamCallback,
    MyCallback_G39DDC_DDC2PreprocessedStreamCallback,
    MyCallback_G39DDC_IFCallback,
    MyCallback_G39DDC_SweepingCallback,
    MyCallback_G39DDC_AudioStreamCallback,
    MyCallback_G39DDC_AudioPlaybackStreamCallback
};

//callback object
IG39DDCDeviceCallback MyCallbackObject=
{
    &MyCallbackMethodsTable
};


VOID __stdcall MyCallback_G39DDC_DDC1StreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,CONST VOID *Buffer,UINT32 NumberOfSamples,UINT32 BitsPerSample)
{
    //it does nothing in this example
}

BOOL __stdcall MyCallback_G39DDC_DDC1PlaybackStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,VOID *Buffer,UINT32 NumberOfSamples,UINT32 BitsPerSample)
{
    //it does nothing in this example
    return FALSE;
}

VOID __stdcall MyCallback_G39DDC_DDC2StreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,CONST float *Buffer,UINT32 NumberOfSamples)
{
    //it does nothing in this example
}

VOID __stdcall MyCallback_G39DDC_DDC2PreprocessedStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,CONST float *Buffer,UINT32 NumberOfSamples,float SlevelPeak,float SlevelRMS)
{
 double Slevel_dBm;

    //convert slevel RMS in Volts to dBm
    Slevel_dBm=10.0*log10(SlevelRMS*SlevelRMS*(1000.0/50.0));

    printf("\rSlevel: RMS[V]=%9.7f V, RMS[dBm]=%6.1f dBm, Peak[V]=%9.7f V",SlevelRMS,Slevel_dBm,SlevelPeak);
}

VOID __stdcall MyCallback_G39DDC_IFCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,CONST SHORT *Buffer,UINT32 NumberOfSamples,UINT32 CenterFrequency,WORD MaxADCAmplitude,UINT32 ADCSampleRate)
{
    //it does nothing in this example
}

BOOL __stdcall MyCallback_G39DDC_SweepingCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,CONST SHORT *Buffer,UINT32 NumberOfSamples,UINT32 CenterFrequency,UINT64 FrontEndFrequency)
{
    //it does nothing in this example
    return FALSE;
}

VOID __stdcall MyCallback_G39DDC_AudioStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,UINT32 Type,CONST float *Buffer,UINT32 NumberOfSamples)
{
 CONST FLOAT *Input;
 WAVEHDR *WaveHdr;
 SHORT *Output;

    if(Type==2)
    {
        //we only want to play audio buffer of type 2 - final audio samples (audio gain, audio filter and volume are applied)


        //alocate memory for waveout stereo audio buffer and WAVEHDR structure
        WaveHdr=(WAVEHDR*)VirtualAlloc(NULL,sizeof(WAVEHDR)+NumberOfSamples*2*sizeof(SHORT),MEM_COMMIT,PAGE_READWRITE);

        ZeroMemory(WaveHdr,sizeof(WAVEHDR));

        WaveHdr->lpData=(LPSTR)(WaveHdr+1); //audio buffer starts after WAVEHDR structure
        WaveHdr->dwBufferLength=NumberOfSamples*2*sizeof(SHORT);

        //retype Buffer to INT32, because bits per sample of DDC type 1 (specified by SetDDCType) is 32 - signed
        Input=(CONST FLOAT*)Buffer;

        //retype audio buffer ptr to signed 16bit
        Output=(SHORT*)WaveHdr->lpData;

        //convert samples from 32bit IEEE float to 16bit integer
        while(NumberOfSamples--)
        {
            *Output++=(SHORT)((*Input++)*32767.0); //left
            *Output++=(SHORT)((*Input++)*32767.0); //right
        }

        waveOutPrepareHeader(hWaveOut,WaveHdr,sizeof(WAVEHDR));

        //increase number of buffers sent to waveout
        InterlockedIncrement(&BufferCount);

        if(waveOutWrite(hWaveOut,WaveHdr,sizeof(WAVEHDR))!=MMSYSERR_NOERROR)
        {
            //decrease number of buffers sent to waveout because an error
            InterlockedDecrement(&BufferCount);
            VirtualFree(WaveHdr,0,MEM_RELEASE);
        }
    }
}

BOOL __stdcall MyCallback_G39DDC_AudioPlaybackStreamCallback(IG39DDCDeviceCallback *Callback,IG39DDCDevice *Device,UINT32 Channel,float *Buffer,UINT32 NumberOfSamples)
{
    //it does nothing in this example
    return FALSE;
}



//procedure of thread used for freeing audio buffers returned to application
//from audio subsystem
DWORD __stdcall WaveOutThreadProcedure(void *Param)
{
 MSG Msg;
 WAVEHDR *WaveHdr;

    while(GetMessage(&Msg,NULL,0,0))
    {
        switch(Msg.message)
        {
            case MSG_EXITTHREAD:
                //exit thread
                return 0;

            case MM_WOM_DONE:
                WaveHdr=(WAVEHDR*)Msg.lParam;

                waveOutUnprepareHeader(hWaveOut,WaveHdr,sizeof(WAVEHDR));

                VirtualFree(WaveHdr,0,MEM_RELEASE);

                InterlockedDecrement(&BufferCount);
                break;

            default:
                DispatchMessage(&Msg);
                break;
        }
    }

    return 0;
}

BOOL Initialize(void)
{
 WAVEFORMATEX WaveFormat;

    hAPI=LoadLibrary("G39DDCAPI.dll");

    if(hAPI==NULL)
    {
        printf("Failed to load G39DDCAPI.dll.\n");
        return FALSE;
    }

    //retrieve addresse of CreateInstance API functions
    CreateInstance=(G39DDCAPI_CREATE_INSTANCE)GetProcAddress(hAPI,"CreateInstance");

    //create instance of device object
    if(!CreateInstance(G39DDC_CLASS_ID_DEVICE,(void**)&Device))
    {
        printf("Failed to create device object. Error code=%08X\n",GetLastError());
        FreeLibrary(hAPI);
        return FALSE;
    }

    //open the first available G39DDC device
    if(!Device->lpVtbl->Open(Device,G39DDC_OPEN_FIRST))
    {
        printf("No available G39DDC device found.\n");
        Device->lpVtbl->Release(Device);
        FreeLibrary(hAPI);
        return FALSE;
    }


    //create thread which frees audio buffers returned to application
    hThread=CreateThread(NULL,0,WaveOutThreadProcedure,NULL,0,&ThreadId);

    if(hThread==NULL)
    {
        printf("Failed to create thread. Error=%u\n",GetLastError());
        Device->lpVtbl->Release(Device);
        FreeLibrary(hAPI);
        return FALSE;
    }

    //initialize waveout
    WaveFormat.wFormatTag=WAVE_FORMAT_PCM;
    WaveFormat.nChannels=2;
    WaveFormat.nSamplesPerSec=48000;
    WaveFormat.wBitsPerSample=16;
    WaveFormat.nBlockAlign=WaveFormat.nChannels*WaveFormat.wBitsPerSample/8;
    WaveFormat.nAvgBytesPerSec=WaveFormat.nBlockAlign*WaveFormat.nSamplesPerSec;
    WaveFormat.cbSize=0;

    //reset buffer count
    BufferCount=0;

    if(waveOutOpen(&hWaveOut,WAVE_MAPPER,&WaveFormat,ThreadId,NULL,CALLBACK_THREAD)!=MMSYSERR_NOERROR)
    {
        printf("Failed to open waveout\n");
        PostThreadMessage(ThreadId,MSG_EXITTHREAD,0,0);
        WaitForSingleObject(hThread,INFINITE);
        CloseHandle(hThread);
        Device->lpVtbl->Release(Device);
        FreeLibrary(hAPI);
        return FALSE;
    }

    return TRUE;
}


int main(int argc, char* argv[])
{
 G39DDC_DDC_INFO DDCInfo;
 UINT32 DDCTypeIndex;
 G39DDC_CALLBACKS Callbacks;
 UINT64 Frequency;
 UINT64 FrontEndFrequency;
 INT32 DDC1Frequency;
 INT32 DDC2Frequency;
 INT32 DemodulatorFrequency;

    if(!Initialize())
    {
        return -1;
    }

    //set power
    Device->lpVtbl->SetPower(Device,TRUE);

    //set DDC type of DDC1 to 1 (DDC Bandwidth=24000, SampleRate=32kHz, BitsPerSample=32)
    Device->lpVtbl->SetDDC1(Device,0,1);

    //retrieve current DDC type of the DDC1
    Device->lpVtbl->GetDDC1(Device,0,&DDCTypeIndex);
    Device->lpVtbl->GetDDCInfo(Device,DDCTypeIndex,&DDCInfo);
    printf("DDC type of DDC1: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);

    
    //set DDC type of DDC2 to 1 (DDC Bandwidth=24000, SampleRate=32kHz, BitsPerSample=32)
    Device->lpVtbl->SetDDC2(Device,0,1);

    //retrieve current DDC type of the DDC2
    Device->lpVtbl->GetDDC2(Device,0,&DDCTypeIndex);
    Device->lpVtbl->GetDDCInfo(Device,DDCTypeIndex,&DDCInfo);
    printf("DDC type of DDC2: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);



    //register callback object
    Device->lpVtbl->SetCallback(Device,&MyCallbackObject);

    //tune channel 0 to an AM station frequency
    Device->lpVtbl->SetFrequency(Device,0,702000);

    //get absolute frequency where the demodulator of channel 0 is tuned
    Device->lpVtbl->GetFrequency(Device,0,&Frequency);

    //get front-end frequency
    Device->lpVtbl->GetFrontEndFrequency(Device,&FrontEndFrequency);

    //get DDC1 frequency
    Device->lpVtbl->GetDDC1Frequency(Device,0,&DDC1Frequency);

    //get relative DDC2 frequency for channel 0
    Device->lpVtbl->GetDDC2Frequency(Device,0,&DDC2Frequency);

    //get relative demodulator frequency for channel 0
    Device->lpVtbl->GetDemodulatorFrequency(Device,0,&DemodulatorFrequency);

    printf("Listening to (demodulator absolute frequency): %I64u Hz\n",Frequency);
    printf("Front-end frequency: %I64u Hz\n",FrontEndFrequency);
    printf("DDC1 frequency (relative): %d Hz\n",DDC1Frequency);
    printf("DDC2 frequency (relative): %d Hz\n",DDC2Frequency);
    printf("Demodulator frequency (relative): %d Hz\n",DemodulatorFrequency);


    //set AGC params for channel 0, attack=0.02s, decay=0.5s, ref. level=-15dB
    Device->lpVtbl->SetAGCParams(Device,0,0.020,0.500,-15.0);

    //enable AGC for channel 0
    Device->lpVtbl->SetAGC(Device,0,TRUE);

    //set AM mode for channel 0
    Device->lpVtbl->SetDemodulatorMode(Device,0,G39DDC_MODE_AM);

    //set demodulator filter bandwidth for channel 0 to 8kHz
    Device->lpVtbl->SetDemodulatorFilterBandwidth(Device,0,8000);

    //start DDC1 streaming which has to be running before StartDDC2 is called
    Device->lpVtbl->StartDDC1(Device,0,1024);

    printf("\nPress ENTER to exit.\n\n");

    //start DDC2 streaming for channel 0, it has to be running before StartAudio is called
    Device->lpVtbl->StartDDC2(Device,0,1024);


    //start audio streaming for channel 0, with 1024 samples per buffer
    Device->lpVtbl->StartAudio(Device,0,1024);

    getchar();

    //stop audio streaming for channel 0
    Device->lpVtbl->StopAudio(Device,0);

    //stop DDC2 streaming for channel 0
    Device->lpVtbl->StopDDC2(Device,0);

    //stop DDC1 streaming for channel 0
    Device->lpVtbl->StopDDC1(Device,0);

    //in this case it is not necessary to use StopAudio and StopDDC2
    //because StopDDC1 stop audio and DDC2 streaming too


    waveOutReset(hWaveOut);

    //wait for all buffers sent to waveout are returned
    while(InterlockedCompareExchange(&BufferCount,0,0)!=0)
    {
        Sleep(10);
    }

    PostThreadMessage(ThreadId,MSG_EXITTHREAD,0,0);
    WaitForSingleObject(hThread,INFINITE);
    CloseHandle(hThread);

    waveOutClose(hWaveOut);

    //unregister callback object
    Device->lpVtbl->SetCallback(Device,NULL);

    Device->lpVtbl->SetPower(Device,FALSE);

    Device->lpVtbl->Release(Device);

    FreeLibrary(hAPI);

    return 0;
}